"""ScoreTree package.

This package contains the ScoreTree class and its dependencies. It is used to
generate a score tree from a list of Score and/or ScoreArea instances,
providing with an easy to use, multi-level grade weighting system.

Author:
    Paulo Sanchez (@erlete)
"""


from .core.scores import Score, ScoreArea
from .core.tree import ScoreTree



import pytest
from colorama import Fore, Style

from ..core.scores import Score, ScoreArea


class TestScore:

    def test_init(self):
        Score("", 0, (0, 1))
        Score("", 0, (0, 1), 0)
        Score("", 0, (0, 1), 0)
        Score("", 0, (0, 1), 0, False)

    def test_properties(self):
        score = Score("test", 0, (0, 1))

        assert score.name == "test"
        assert score.weight == 0
        assert score.score_range == (0, 1)
        assert score.value == 0
        assert not score.inverse

        score.name = "test2"
        with pytest.raises(TypeError):
            score.name = 0
            score.name = 0.0
            score.name = None
            score.name = []

        score.weight = 1
        score.weight = 1.0
        with pytest.raises(TypeError):
            score.weight = "0"
            score.weight = None
            score.weight = []

        score.score_range = (0, 1)
        score.score_range = (0.0, 1.0)
        with pytest.raises(TypeError):
            score.score_range = "0"
            score.score_range = None
            score.score_range = []
            score.score_range = [0, 0]

        with pytest.raises(TypeError):
            score.score_range = (0, "a")
            score.score_range = ("a", 0)
            score.score_range = ("1", "0")

        with pytest.raises(ValueError):
            score.score_range = ()
            score.score_range = (0,)
            score.score_range = (0, 0, 0)

        with pytest.raises(ValueError):
            score.score_range = (0, 0)
            score.score_range = (1, 0)

        score.value = 1
        score.value = 1.0
        with pytest.raises(TypeError):
            score.value = "0"
            score.value = None
            score.value = []

        score.inverse = True
        score.inverse = False
        with pytest.raises(TypeError):
            score.inverse = 0
            score.inverse = 0.0
            score.inverse = "0"
            score.inverse = None
            score.inverse = []
            score.inverse = ()

        score.score_range = (0, 1)
        score.value = 0.5
        assert score.score == 0.5

    def test_representation(self):
        score = Score("Test", 0, (0, 1))

        assert repr(score) == "Score(Test)"
        assert str(score) == "Test (0.00%): 0.00%"

        assert score._render(0) == (
            f"{Fore.RED}{Style.DIM}Test (0.00%): 0.00%"
        )
        assert score._render(1) == score._render() == (
            f"{Fore.RED}{Style.DIM}└── Test (0.00%): 0.00%"
        )

        with pytest.raises(TypeError):
            score._render("0")
            score._render(None)
            score._render([])
            score._render(())


class TestScoreArea:

    def test_init(self):
        ScoreArea("", 0, [])
        ScoreArea("", 0, [Score("", 0, (0, 1))])

    def test_properties(self):
        score_area = ScoreArea("test", 0, [])

        assert score_area.name == "test"
        assert score_area.weight == 0
        assert score_area.items == []

        score_area.name = "test2"
        with pytest.raises(TypeError):
            score_area.name = 0
            score_area.name = 0.0
            score_area.name = None
            score_area.name = []

        score_area.weight = 1
        score_area.weight = 1.0
        with pytest.raises(TypeError):
            score_area.weight = "0"
            score_area.weight = None
            score_area.weight = []

        score_area.items = []
        score_area.items = [Score("", 0, (0, 1))]
        score_area.items = [Score("", 0, (0, 1)), Score("", 0, (0, 1))]
        score_area.items = [ScoreArea("", 0, [])]
        score_area.items = [ScoreArea("", 0, []), ScoreArea("", 0, [])]
        score_area.items = [Score("", 0, (0, 1)), ScoreArea("", 0, [])]
        with pytest.raises(TypeError):
            score_area.items = 0
            score_area.items = 0.0
            score_area.items = ""
            score_area.items = None
            score_area.items = ()

        with pytest.raises(TypeError):
            score_area.items = [0]
            score_area.items = [0.0]
            score_area.items = [""]
            score_area.items = [None]
            score_area.items = [()]

        score_area.weight = 1
        score_area.items = [Score("", 1, (0, 1), 1)]

        assert score_area.score == 1

    def test_representation(self):
        score_area = ScoreArea("Test", 1, [Score("Test1", 1, (0, 1), 1)])

        assert repr(score_area) == "ScoreArea(Test)"
        assert str(score_area) == (
            "Test (100.00%): 100.00%\n    Test1 (100.00%): 100.00%"
        )

        assert score_area._render(0) == (
            f"{Fore.GREEN}{Style.NORMAL}Test (100.00%): 100.00%\n"
            + f"{Fore.GREEN}{Style.DIM}└── Test1 (100.00%): 100.00%"
        )
        assert score_area._render(1) == score_area._render() == (
            f"{Fore.GREEN}{Style.NORMAL}└── Test (100.00%): 100.00%\n"
            + f"{Fore.GREEN}{Style.DIM}    └── Test1 (100.00%): 100.00%"
        )

        with pytest.raises(TypeError):
            score_area._render("0")
            score_area._render(None)
            score_area._render([])
            score_area._render(())



"""Formatting module.

This module contains several text formatting utilities for indentation and
colorization.

Author:
    Paulo Sanchez (@erlete)
"""


from colorama import Fore, init

init()


class Formatter:
    """Formatting class.

    This class contains several methods used for string manipulation and
    formatting operations, such as indentation or colorization.

    Attributes:
        START_CHAR (str): character used to indicate start of a new line.
        MIDDLE_CHAR (str): character used to indicate continuation of a line.
        COLOR_ENABLED (bool): whether colorization is enabled or not.
    """

    START_CHAR = "└"
    MIDDLE_CHAR = "─"
    COLOR_ENABLED = True

    @classmethod
    def indent(cls, count: int = 1) -> str:
        """Generate an indentation string.

        Args:
            count (int, optional): number of indents. Defaults to 1.

        Returns:
            str: indented string.

        Raises:
            TypeError: if count is not an integer.
        """
        if not isinstance(count, int):
            raise TypeError("count must be an integer.")

        count = max(0, count)  # Value normalization.

        if count == 0:
            return ""

        if count == 1:
            return (
                f"{cls.START_CHAR}{cls.MIDDLE_CHAR * 2} "
            )

        return (
            " " * 4 * (count - 1)
            + f"{cls.START_CHAR}{cls.MIDDLE_CHAR * 2} "
        )

    @classmethod
    def colorize(cls, text: str, value: int | float) -> str:
        """Colorize text based on a value.

        Args:
            text (str): text to colorize.
            value (int | float): value to base colorization from 0 to 1 (both
                included).

        Returns:
            str: colorized text.

        Raises:
            TypeError: if text is not a string.
            TypeError: if value is not an integer or float.
        """
        if not isinstance(text, str):
            raise TypeError("text must be a string.")

        if not isinstance(value, (int, float)):
            raise TypeError("value must be an integer or float.")

        value = min(1, max(0, value))  # Value normalization.

        return text if not cls.COLOR_ENABLED else {
            value < .5: Fore.RED,
            .5 <= value <= .75: Fore.YELLOW,
            .75 < value: Fore.GREEN
        }[True] + text



import pytest
from colorama import Fore, Style

from ..core.scores import Score, ScoreArea
from ..core.tree import ScoreTree


class TestScoreTree:

    def test_init(self):
        ScoreTree([Score("test", 1, (0, 1))])
        ScoreTree([Score("test", 1, (0, 1))], False)

    def test_properties(self):
        score = Score("test", 1, (0, 1), 1)
        score_tree = ScoreTree([score])

        assert score_tree.items == [score]
        assert score_tree.colorized

        score_tree.items = [Score("test", 1, (0, 1))]
        score_tree.items = [
            Score("test", 0.5, (0, 1)),
            Score("test2", 0.5, (0, 1))
        ]
        score_tree.items = [
            ScoreArea("test", 1, [
                Score("test", 0.5, (0, 1)),
                Score("test2", 0.5, (0, 1))
            ])
        ]
        score_tree.items = [
            ScoreArea("test", 0.5, [
                Score("test", 0.5, (0, 1)),
                Score("test2", 0.5, (0, 1))
            ]),
            Score("test", 0.25, (0, 1)),
            Score("test2", 0.25, (0, 1))
        ]
        with pytest.raises(TypeError):
            score_tree.items = None
            score_tree.items = 0
            score_tree.items = 0.0
            score_tree.items = ""
            score_tree.items = ()

        with pytest.raises(TypeError):
            score_tree.items = [None]
            score_tree.items = [0]
            score_tree.items = [0.0]
            score_tree.items = [""]
            score_tree.items = [()]

        score_tree.colorized = True
        score_tree.colorized = False
        with pytest.raises(TypeError):
            score_tree.colorized = None
            score_tree.colorized = 0
            score_tree.colorized = 0.0
            score_tree.colorized = ""
            score_tree.colorized = ()

        score_tree.items = [Score("test", 1, (0, 1), 0.5)]
        assert score_tree.score == 0.5

    def test_check_score(self):
        with pytest.raises(ValueError):
            ScoreTree([])
            ScoreTree([Score("test", 0, (0, 1))])
            ScoreTree([Score("test", .5, (0, 1))])
            ScoreTree([Score("test", .99, (0, 1))])
            ScoreTree([Score("test", 1.01, (0, 1))])
            ScoreTree([Score("test", -1, (0, 1))])

        with pytest.raises(ValueError):
            ScoreTree([ScoreArea("test", 0, [])])
            ScoreTree([ScoreArea("test", .5, [])])
            ScoreTree([ScoreArea("test", .99, [])])
            ScoreTree([ScoreArea("test", 1.01, [])])
            ScoreTree([ScoreArea("test", -1, [])])

    def test_representation(self):
        score_tree = ScoreTree([
            Score("Test", 0.5, (0, 1)),
            Score("Test2", 0.5, (0, 1), .5)
        ])

        assert repr(score_tree) == "<ScoreTree with 2 items>"
        assert str(score_tree) == (
            f"{Fore.RED}{Fore.RED}{Style.DIM}Test (50.00%): 0.00%"
            + f"{Style.RESET_ALL}\n{Fore.YELLOW}{Fore.YELLOW}{Style.DIM}"
            + f"Test2 (50.00%): 50.00%{Style.RESET_ALL}"
        )



"""Track performance score example (intermediate).

This example explores the nesting capabilities of the ScoreArea class in
combination with the Score class, allowing complex score trees to be created.

Author:
    Paulo Sanchez (@erlete)
"""

from scoretree import Score, ScoreArea, ScoreTree

st = ScoreTree([
    ScoreArea("Track 1", 1, [
        ScoreArea("Dynamics", 0.6, [
            Score("Top speed (m/s)", 0.5, (0, 100), 88.2),
            Score("Elapsed time (s)", 0.3, (20, 60), 31.2, True),
            Score("Traveled distance (m)", 0.2, (250, 785), 327.12, True)
        ]),
        ScoreArea("Efficiency", 0.4, [
            Score("Fuel consumption (l)", 0.72, (39.13, 69.32), 58.12, True),
            ScoreArea("Energy consumption", 0.28, [
                Score("Battery consumption (kWh)", 0.65, (0, 43.74), 6, True),
                Score("Regenerative braking (kWh)", 0.35, (0, 16.1), 8.16)
            ])
        ])
    ])
], colorized=True)

print(f"{' Score tree ':=^80}\n{st}")
print(f"{f' Total track score: {st.score * 100:.2f} ':=^80}")



"""Score tree module.

This module contains the ScoreTree class, which is used to generate a score
tree from a list of Score and/or ScoreArea instances.

Author:
    Paulo Sanchez (@erlete)
"""


from __future__ import annotations

from colorama import Style

from .formatter import Formatter
from .scores import Score, ScoreArea


class ScoreTree(Formatter):
    """Score tree generation class.

    This class serves as interface for score tree generation. By providing the
    items attribute with a list of Score and/or ScoreArea instances (last ones
    being nestable), all dependent scores will be computed and displayed with
    ease.

    Attributes:
        items (list[Score | ScoreArea]): list of Score or ScoreArea items.
        score (float): weighted score.
        colorized (bool): whether colorization is enabled or not.
    """

    def __init__(
        self,
        items: list[Score | ScoreArea],
        colorized: bool = True
    ) -> None:
        """Initialize a ScoreTree instance.

        Args:
            items (list[Score | ScoreArea]): list of Score or ScoreArea items.
            colorized (bool, optional): whether colorization is enabled or not.
        """
        self.items = items
        self.colorized = colorized

    @property
    def items(self) -> list[Score | ScoreArea]:
        """Get score items.

        Returns:
            list[Score | ScoreArea]: score items.
        """
        return self._items

    @items.setter
    def items(self, value: list[Score | ScoreArea]) -> None:
        """Set score items.

        Args:
            value (list[Score | ScoreArea]): score items.

        Raises:
            TypeError: if value is not a list.
            TypeError: if value contains elements that are not Score or
                ScoreArea instances.
        """
        if not isinstance(value, list):
            raise TypeError(
                "expected type list[Score | ScoreArea] for"
                + f" {self.__class__.__name__}.items but got"
                + f" {type(value).__name__} instead"
            )

        if not all(isinstance(item, (Score, ScoreArea)) for item in value):
            raise TypeError(
                "expected type list[Score | ScoreArea] for"
                + f" {self.__class__.__name__}.items elements but got"
                + f" {type(value).__name__} instead"
            )

        self._items = value

        # Weight completeness self-checking:
        self.check_weights(self)

    @property
    def colorized(self) -> bool:
        """Get colorization flag.

        Returns:
            bool: colorization flag.
        """
        return self._colorized

    @colorized.setter
    def colorized(self, value: bool) -> None:
        """Set colorization flag.

        Args:
            value (bool): colorization flag.

        Raises:
            TypeError: if value is not a bool.
        """
        if not isinstance(value, bool):
            raise TypeError(
                "expected type bool for"
                + f" {self.__class__.__name__}.colorized but got"
                + f" {type(value).__name__} instead"
            )

        self._colorized = value
        Formatter.COLOR_ENABLED = value

    @property
    def score(self) -> float:
        """Get weighted score.

        Returns:
            float: weighted score.
        """
        return sum(level.score * level.weight for level in self.items)

    @classmethod
    def check_weights(cls, score_collection: ScoreArea | ScoreTree) -> None:
        """Check if weights of a ScoreArea or ScoreTree add up to 1.

        Args:
            score_collection (ScoreArea | ScoreTree): score area or score tree
                to check.

        Raises:
            ValueError: if weights do not add up to 1.
        """
        # Accumulation:
        total = 0.0
        for item in score_collection.items:
            total += item.weight

            # Recursive checking:
            if isinstance(item, ScoreArea):
                cls.check_weights(item)

        # Score area checking:
        if total != 1 and isinstance(score_collection, ScoreArea):
            raise ValueError(
                f"\"{score_collection.name}\""
                + f" score weights do not add up to 1 ({total})"
            )

        # Score tree checking:
        elif total != 1 and isinstance(score_collection, ScoreTree):
            raise ValueError(
                f"score tree weights do not add up to 1 ({total})"
            )

    def __repr__(self) -> str:
        """Get short string representation of the score tree.

        Returns:
            str: short string representation of the score tree.
        """
        return f"<ScoreTree with {len(self.items)} items>"

    def __str__(self) -> str:
        """Get long representation of the score tree.

        Returns:
            str: long representation of the score tree.
        """
        return "\n".join(
            self.colorize(
                f"{item._render(0)}{Style.RESET_ALL}",
                item.score
            ) for item in self.items
        )



import pytest
from colorama import Fore

from ..core.formatter import Formatter


class TestFormatter:

    def test_indent(self):
        with pytest.raises(TypeError):
            Formatter.indent(1.1)
            Formatter.indent("1")
            Formatter.indent(True)

        assert Formatter.indent(-1) == ""
        assert Formatter.indent(0) == ""
        assert Formatter.indent(1) == (
            f"{Formatter.START_CHAR}{Formatter.MIDDLE_CHAR * 2} "
        )
        assert Formatter.indent(2) == (
            f"{' ' * 4}{Formatter.START_CHAR}{Formatter.MIDDLE_CHAR * 2} "
        )

    def test_colorize(self):
        with pytest.raises(TypeError):
            Formatter.colorize(1, 0)
            Formatter.colorize(1.1, 0)
            Formatter.colorize([1, 2, 3], 0)

        with pytest.raises(TypeError):
            Formatter.colorize("", "1")
            Formatter.colorize("", [1, 2, 3])

        assert Formatter.colorize("red", 0) == f"{Fore.RED}red"
        assert Formatter.colorize("red", 0.49) == f"{Fore.RED}red"
        assert Formatter.colorize("yellow", 0.5) == f"{Fore.YELLOW}yellow"
        assert Formatter.colorize("yellow", 0.75) == f"{Fore.YELLOW}yellow"
        assert Formatter.colorize("green", 0.76) == f"{Fore.GREEN}green"
        assert Formatter.colorize("green", 1) == f"{Fore.GREEN}green"



"""Track performance score example (beginner).

This example introduces the ScoreArea class, which allows for the creation of
nested score areas, which in turn can contain other score areas and/or scores.

Author:
    Paulo Sanchez (@erlete)
"""

from scoretree import Score, ScoreArea, ScoreTree

st = ScoreTree([
    ScoreArea(name="Track 1", weight=0.4, items=[
        Score("Top speed (m/s)", 0.5, (0, 100), 48.12),
        Score("Elapsed time (s)", 0.3, (20, 60), 31.2, True),
        Score("Traveled distance (m)", 0.2, (250, 785), 327.12, True)
    ]),
    ScoreArea(name="Track 2", weight=.6, items=[
        Score("Top speed (m/s)", 0.5, (0, 100), 72.32),
        Score("Elapsed time (s)", 0.3, (20, 60), 26.2, True),
        Score("Traveled distance (m)", 0.2, (250, 785), 295.12, True)
    ])
], colorized=True)

print(f"{' Score tree ':=^80}\n{st}")
print(f"{f' Total simulation score: {st.score * 100:.2f} ':=^80}")



"""Track performance score example (advanced).

This example shows how to iteratively generate a score tree from a original
data source (i.e. a dictionary containing track results).

Author:
    Paulo Sanchez (@erlete)
"""

from scoretree import Score, ScoreArea, ScoreTree

data = {
    "Track 1": {
        "top_speed": 88.2,
        "elapsed_time": 31.2,
        "traveled_distance": 327.12,
        "fuel_consumption": 58.12,
        "battery_consumption": 6,
        "regenerative_braking": 8.16
    },
    "Track 2": {
        "top_speed": 72.32,
        "elapsed_time": 26.2,
        "traveled_distance": 295.12,
        "fuel_consumption": 48.12,
        "battery_consumption": 4.5,
        "regenerative_braking": 6.12
    },
    "Track 3": {
        "top_speed": 92.12,
        "elapsed_time": 28.2,
        "traveled_distance": 315.12,
        "fuel_consumption": 52.12,
        "battery_consumption": 5.5,
        "regenerative_braking": 7.12
    }
}

track_weights = {
    "Track 1": 0.3,
    "Track 2": 0.5,
    "Track 3": 0.2
}

st = ScoreTree([
    ScoreArea(track, weight, [
        ScoreArea("Dynamics", 0.6, [
            Score("Top speed (m/s)", 0.5, (0, 100), results["top_speed"]),
            Score("Elapsed time (s)", 0.3, (20, 60),
                  results["elapsed_time"], True),
            Score("Traveled distance (m)", 0.2, (250, 785),
                  results["traveled_distance"], True)
        ]),
        ScoreArea("Efficiency", 0.4, [
            Score("Fuel consumption (l)", 0.72, (39.13, 69.32),
                  results["fuel_consumption"], True),
            ScoreArea("Energy consumption", 0.28, [
                Score("Battery consumption (kWh)", 0.65, (0, 43.74),
                      results["battery_consumption"], True),
                Score("Regenerative braking (kWh)", 0.35, (0, 16.1),
                      results["regenerative_braking"])
            ])
        ])
    ])
    for (track, results), weight in zip(data.items(), track_weights.values())
], colorized=True)

print(f"{' Score tree ':=^80}\n{st}")
print(f"{f' Total simulation score: {st.score * 100:.2f} ':=^80}")



"""Track performance score example (beginner).

This example shows how to use the most basic features of the scoretree module,
which is the generation of a score list (not actually a tree, due to its
simplicity) from a list of Score instances.

Author:
    Paulo Sanchez (@erlete)
"""

from scoretree import Score, ScoreTree

st = ScoreTree([
    # Top speed score that weights 60% of the total score, can have any value
    #   between 0 and 100, and has a current value of 48.12. The higher the
    #   top speed, the better the score.
    Score(
        name="Top speed (m/s)",
        weight=0.5,
        score_range=(0, 100),
        value=48.12
    ),
    # Elapsed time score. This time, the score is inverted, since the lower
    #   the time, the better the score.
    Score(
        name="Elapsed time (s)",
        weight=0.3,
        score_range=(20, 60),
        value=31.2,
        inverse=True
    ),
    Score(
        name="Traveled distance (m)",
        weight=0.2,
        score_range=(250, 785),
        value=327.12,
        inverse=True
    )
], colorized=True)

print(f"{' Score list ':=^80}\n{st}")
print(f"{f' Total track score: {st.score * 100:.2f} ':=^80}")



"""Score classes module.

This module contains the Score and ScoreArea classes, which are used to
represent a score and a score area, respectively. The Score class is the
smallest score unit, containing a name, a weight for ponderation, a score range
that defines the evaluation range and a value. The ScoreArea class is a score
area, which contains a name, a weight for ponderation and a list of scores or
score areas.

Author:
    Paulo Sanchez (@erlete)
"""


from __future__ import annotations

from colorama import Style

from .formatter import Formatter


class Score(Formatter):
    """Minimal score representation unit.

    This class represents the smallest score unit, which contains a name, a
    weight for ponderation, a score range that defines the evaluation range and
    a value. Furthermore, the computation can be inverted, for example, if the
    score range is from 0 to 10, a value of 10 will be 0 and a value of 0 will
    be 1. This is useful for scores that are better when lower, like distance
    to end.

    Attributes:
        name (str): score name.
        weight (float): score weight.
        score_range (tuple[float, float]): score range from
            minimum to maximum.
        value (float, optional): current value. Defaults to 0.
        inverse (bool, optional): whether to invert the score calculation
            process. Defaults to False.
        score (float): score.
    """

    def __init__(
        self,
        name: str,
        weight: int | float,
        score_range: tuple[int | float, int | float],
        value: int | float = 0,
        inverse: bool = False
    ) -> None:
        """Initialize a Score instance.

        Args:
            name (str): score name.
            weight (int | float): score weight.
            score_range (tuple[int | float, int | float]): score range from
                minimum to maximum.
            value (int | float, optional): current value. Defaults to 0.
            inverse (bool, optional): whether to invert the score calculation
                process. Defaults to False.
        """
        self.name = name
        self.weight = weight
        self.score_range = score_range
        self.value = value
        self.inverse = inverse

    @property
    def name(self) -> str:
        """Get score name.

        Returns:
            str: score name.
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set score name.

        Args:
            value (str): score name.

        Raises:
            TypeError: if value is not a string.
        """
        if not isinstance(value, str):
            raise TypeError(
                "expected type str for"
                + f" {self.__class__.__name__}.name but got"
                + f" {type(value).__name__} instead"
            )

        self._name = value

    @property
    def weight(self) -> float:
        """Get score weight.

        Returns:
            float: score weight.
        """
        return self._weight

    @weight.setter
    def weight(self, value: int | float) -> None:
        """Set score weight.

        Args:
            value (int | float): score weight.

        Raises:
            TypeError: if value is not an int or float.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.weight but got"
                + f" {type(value).__name__} instead"
            )

        self._weight = float(value)

    @property
    def score_range(self) -> tuple[float, float]:
        """Get score range.

        Returns:
            tuple[float, float]: score range.
        """
        return self._score_range

    @score_range.setter
    def score_range(self, value: tuple[int | float, int | float]) -> None:
        """Set score range.

        Args:
            value (tuple[int | float, int | float]): score range.

        Raises:
            TypeError: if value is not a tuple.
            TypeError: if value elements are not int or float.
            ValueError: if value length is not 2.
        """
        if not isinstance(value, tuple):
            raise TypeError(
                "expected type tuple[int | float, int | float] for"
                + f" {self.__class__.__name__}.score_range but got"
                + f" {type(value).__name__} instead"
            )

        if not all(isinstance(item, (int, float)) for item in value):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.score_range elements but got"
                + f" {type(value).__name__} instead"
            )

        if not len(value) == 2:
            raise ValueError(
                "expected length 2 for"
                + f" {self.__class__.__name__}.score_range but got"
                + f" {len(value)} instead"
            )

        if not value[0] < value[1]:
            raise ValueError(
                "expected first element to be less than second element for"
                + f" {self.__class__.__name__}.score_range but got"
                + f" {value[0]} >= {value[1]} instead"
            )

        self._score_range = (float(value[0]), float(value[1]))

    @property
    def value(self) -> float:
        """Get current value.

        Returns:
            float: current value.
        """
        return self._value

    @value.setter
    def value(self, value: int | float) -> None:
        """Set current value.

        Args:
            value (int | float): current value.

        Raises:
            TypeError: if value is not an int or float.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.value but got"
                + f" {type(value).__name__} instead"
            )

        self._value = float(value)

    @property
    def inverse(self) -> bool:
        """Get inverse operation flag.

        Returns:
            bool: inverse operation flag.
        """
        return self._inverse

    @inverse.setter
    def inverse(self, value: bool) -> None:
        """Set inverse operation flag.

        Args:
            value (bool): inverse operation flag.

        Raises:
            TypeError: if value is not a bool.
        """
        if not isinstance(value, bool):
            raise TypeError(
                "expected type bool for"
                + f" {self.__class__.__name__}.inverse but got"
                + f" {type(value).__name__} instead"
            )

        self._inverse = value

    @property
    def score(self) -> float:
        """Get score.

        Returns:
            float: score.
        """
        # Compute score, invert it (if specified) and normalize it:
        return min(1, max(abs(
            self._inverse - (min(
                self._value - self._score_range[0],
                self._score_range[1] - self._score_range[0]
            ) / (self._score_range[1] - self._score_range[0]))
        ), 0))

    def _render(self, indent: int = 1) -> str:
        """Render formatted score.

        Args:
            indent (int, optional): indentation level. Defaults to 1.

        Returns:
            str: formatted score.

        Raises:
            TypeError: if indent is not an integer.
        """
        if not isinstance(indent, int):
            raise TypeError("indent must be an integer")

        indent = max(0, indent)  # Value normalization.

        return self.colorize(
            f"{Style.DIM}{self.indent(indent)}{self!s}"
            if Formatter.COLOR_ENABLED else f"{self.indent(indent)}{self!s}",
            self.score
        )

    def __repr__(self) -> str:
        """Get short representation of the score.

        Returns:
            str: short representation of the score.
        """
        return f"Score({self.name})"

    def __str__(self) -> str:
        """Get long representation of the score.

        Returns:
            str: long representation of the score.
        """
        return (
            f"{self.name} ({self.weight * 100:.2f}%): {self.score * 100:.2f}%"
        )


class ScoreArea(Formatter):
    """Score area representation unit.

    The score area represents a set of scores or child score areas that can be
    used to group scores by level or category. It contains a name, a weight for
    ponderation and a list of scores or score areas.

    Attributes:
        name (str): score area name.
        weight (float): score area weight.
        items (list[Score | ScoreArea]): score area items (can either be
            Score or ScoreArea instances).
        score (float): weighted score.
    """

    def __init__(
        self,
        name: str,
        weight: int | float,
        items: list[Score | ScoreArea]
    ) -> None:
        """Initialize a ScoreArea instance.

        Args:
            name (str): score area name.
            weight (int | float): score area weight.
            items (list[Score | ScoreArea]): score area items (can either be
                Score or ScoreArea instances).
        """
        self.name = name
        self.weight = weight
        self.items = items

    @property
    def name(self) -> str:
        """Get score area name.

        Returns:
            str: score area name.
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set score area name.

        Args:
            value (str): score area name.

        Raises:
            TypeError: if value is not a string.
        """
        if not isinstance(value, str):
            raise TypeError(
                "expected type str for"
                + f" {self.__class__.__name__}.name but got"
                + f" {type(value).__name__} instead"
            )

        self._name = value

    @property
    def weight(self) -> float:
        """Get score area weight.

        Returns:
            float: score area weight.
        """
        return self._weight

    @weight.setter
    def weight(self, value: int | float) -> None:
        """Set score area weight.

        Args:
            value (int | float): score area weight.

        Raises:
            TypeError: if value is not an int or float.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.weight but got"
                + f" {type(value).__name__} instead"
            )

        self._weight = float(value)

    @property
    def items(self) -> list[Score | ScoreArea]:
        """Get score area items.

        Returns:
            list[Score | ScoreArea]: score area items.
        """
        return self._items

    @items.setter
    def items(self, value: list[Score | ScoreArea]) -> None:
        """Set score area items.

        Args:
            value (list[Score | ScoreArea]): score area items.

        Raises:
            TypeError: if value is not a list.
            TypeError: if value elements are not Score or ScoreArea instances.
        """
        if not isinstance(value, list):
            raise TypeError(
                "expected type list[Score | ScoreArea] for"
                + f" {self.__class__.__name__}.items but got"
                + f" {type(value).__name__} instead"
            )

        if not all(isinstance(item, (Score, ScoreArea)) for item in value):
            raise TypeError(
                "expected type Score | ScoreArea for"
                + f" {self.__class__.__name__}.items elements but got"
                + f" {type(value).__name__} instead"
            )

        self._items = value

    @property
    def score(self) -> float:
        """Get weighted score.

        Returns:
            float: weighted score.
        """
        return sum(item.score * item._weight for item in self.items)

    def _render(self, indent: int = 1) -> str:
        """Render formatted score area.

        Args:
            indent (int, optional): indentation level. Defaults to 1.

        Returns:
            str: formatted score area.

        Raises:
            TypeError: if indent is not an integer.
        """
        if not isinstance(indent, int):
            raise TypeError("indent must be an integer")

        indent = max(0, indent)  # Value normalization.

        return self.colorize(
            (
                Style.NORMAL
                + f"{self.indent(indent)}{self.name}"
                + f" ({self.weight * 100:.2f}%): {self.score * 100:.2f}%\n"
                + "\n".join(
                    item._render(indent + 1)
                    for item in self.items
                )
            ) if Formatter.COLOR_ENABLED else (
                f"{self.indent(indent)}{self.name}"
                + f" ({self.weight * 100:.2f}%): {self.score * 100:.2f}%\n"
                + "\n".join(
                    item._render(indent + 1)
                    for item in self.items
                )
            ), self.score
        )

    def __repr__(self) -> str:
        """Get short representation of the score area.

        Returns:
            str: short representation of the score area.
        """
        return f"ScoreArea({self.name})"

    def __str__(self) -> str:
        """Get long representation of the score area.

        Returns:
            str: long representation of the score area.
        """
        return (
            f"{self.name} ({self.weight * 100:.2f}%):"
            + f" {self.score * 100:.2f}%\n"
            + "\n".join(f"{' ' * 4}{item!s}" for item in self.items)
        )
